//============================================================================
// Name        : hash_multiset.cpp
// Author      : Kamil Halicki H6ISI
// Version     : 1.0
// Copyright   :
// Description : Demonstracja kontenera hash_multiset
//============================================================================

/* hash_multiset czyli wielozbir haszujcy jest to kontener umoliwiajcy przechowywanie elementw ktre nie musz by unikatowe 
 *  oraz zapewniajcy szybkie wyszukiwanie i zliczanie elementw  
 **/ 


#include <iostream>
#include <sstream>
#include <cstring>

/* hash_multiset nie jest wchodzi w skad standardu c++ a jest rozszerzeniem dodanym przez sgi
 * dlatego w rnych kompilatorach moe by zdefiniowany w innym pliku nagwkowym i innej przestrzeni nazw.
 */
#include <ext/hash_set>


using namespace std;



/* W implementacji g++ hash_multiset zdefiniowany jest w przestrzeni nazw __gnu_cxx */
using __gnu_cxx::hash_multiset;
using __gnu_cxx::hash;



/* Implementacja g++ nie zawiera funkcji haszujcej dla typu string ale moemy w atwy sposb rozszerzy bibliotek */
namespace __gnu_cxx {

/* Definiujemy specjalizacj szablonu hash dla typu string */
template<>
    struct hash<string>
    {
      size_t
      operator()(const string &__s) const
      { return __stl_hash_string(__s.c_str()); }  // wykorzystujemy funkcj obliczajc hash dla typu const char *
 };
}


/* definiuj nowy typ dla wielozbioru acuchw znakw dziki czemu skraca si zapis i w atwy sposb mona zmieni np. funkcj haszujc */
typedef hash_multiset<string> strset;



/* Funkcja wypisuje na ekran ile razy sowo what pojawia si w wielozbiorze set */
void ile(const strset& set, const char * what) {
	cout << what << ": " ;

	/* count zwraca liczb elementw o podanym kluczu
	 * what jest typu const char * tak wi tworz tymczasowy obiekt string inicjowany wartoci what
	 */
	cout << set.count(string(what)) << endl;
}

int main() {

	cout << "Hash MultiSet" << endl;


	stringstream lyrics;

	/* To jest tekst ktrego poszczeglne wyrazy bd dodawane do wielozbioru */
	lyrics <<  "and in the sea there is a fish \
				a fish that has a secret wish \
				a wish to be a big cactus \
				with a pink flower on it";



	/* Tutaj tworzony jest nowy obiekt wielozbioru haszowanego */
	strset words;

	/* Obiekt word przechowuje aktualnie przetwarzane sowo */
	string word;

	cout << endl << "Kolejno dodawania: " << endl ;

	while( lyrics.eof() == false ) {
		/* operator strumienniowy przekazuje poszczeglne wyrazy do obiektu word */
		lyrics >> word;


		cout << word << ", ";

		/* W tym miejscu dodajemy sowa do zbioru. Wewntrz kontenera zostanie utworzona kopia obiektu word. */
		words.insert(word);
	}

	cout << endl;
	cout << endl << "Kolejno iterowania" << endl;

	/* Do iterowania po kolekcji wykorzystujemy stay iterator i pierwszy element pobieramy za pomoc metody begin() */
	strset::const_iterator iter = words.begin();

	/* Iterujc sprawdzamy czy nie osignity zosta koniec kolekcji */
	while( iter != words.end() ) {

		cout << *iter << ", ";

		/* Iterator wskazuje na nastpny element w kontenerze */
		++iter;
	}
    cout << endl;

    cout << endl << "Liczba wystpie:" << endl;

    /* Funkcja ile wywietla ile razy dane sowo znajduje si w wielozbiorze */
	ile(words, "fish");
	ile(words, "big");
	ile(words, "wish");
	ile(words, "a");

	return 0;
}


